#-----
#Path On Face - Path process handler
#-----
#version 1.02
#-----

from vcRslProcessHandler import * 
from vcBehaviour import * 
import vcMatrix, vcVector, math 

RAD_TO_DEG = math.pi/180.0 
  
app = getApplication() 
comp = getComponent()
mt = None

#-----
def OnStatementExecute(exe, stat):
  #Execute as process path with possible signals
  global mt
  
  prop_signal = stat.getProperty('ProcessSignal')
  prop_approach = stat.getProperty('ApproachMotionPtp')
  prop_approach_speed = stat.getProperty('ApproachSpeed')
  ctr = exe.Controller
  if not mt:
    mt = ctr.createTarget()
  
  if stat.Base:
    mt.BaseName = stat.Base.Name
  else:
    mt.BaseName = ""
  if stat.Tool:
    mt.ToolName = stat.Tool.Name
  else:
    mt.ToolName = ""
  
  count = len(stat.Positions)
  process_on = False
  
  mt.AccuracyMethod = VC_MOTIONTARGET_AM_DISTANCE
  mt.MotionType = VC_MOTIONTARGET_MT_LINEAR
  if prop_approach.Value:
    mt.MotionType = VC_MOTIONTARGET_MT_JOINT
    mt.JointSpeedFactor = min(prop_approach_speed.Value,100.0) / 100.0
  
  for i in range(count):
    m = stat.getSchemaValue(i, 'Position')
    speed = stat.getSchemaValue(i, 'MaxSpeed')
    process_point = stat.getSchemaValue(i, 'ProcessPoint')
    accuracy = stat.getSchemaValue(i, 'AccuracyValue')
    
    mt.Target = m
    mt.CartesianSpeed = speed
    mt.AccuracyValue = accuracy
    if not process_on and process_point:
      mt.AccuracyValue = 0
    elif process_on and i < count - 1 and not stat.getSchemaValue(i+1, 'ProcessPoint'):
      mt.AccuracyValue = 0
    elif i == count - 1:
      mt.AccuracyValue = 0
    
    if not process_on:
      ctr.addTarget(mt)
      if process_point:
        #Move through via points and set process signal
        ctr.move()
        ctr.clearTargets()
        if prop_signal.Value >= 0:
          exe.DigitalOutputSignals.output(prop_signal.Value, True)
        process_on = True
    else:
      if not process_point:
        #Move through process points and reset process signal
        ctr.move()
        ctr.clearTargets()
        if prop_signal.Value >= 0:
          exe.DigitalOutputSignals.output(prop_signal.Value, False)
        process_on = False
      ctr.addTarget(mt)
      
    mt.MotionType = VC_MOTIONTARGET_MT_LINEAR
    mt.JointSpeedFactor = 100.0
  #endfor
    
  #Move through last set of points and reset process signal
  ctr.move()
  if process_on:
    if prop_signal.Value >= 0:
      exe.DigitalOutputSignals.output(prop_signal.Value, False)
    process_on = False
#-----