#-----
#Path On Face
#
#Add-on for creating process paths on surfaces
#-----
#version 1.03
#-----

from vcCommand import *
import vcVector
import vcMatrix
import os.path
from math import *
from vcHelpers.Application import *

#-------------------------------------------------------------------------------

app = getApplication()
cmd = getCommand() 
cmdPick = app.findCommand('interactiveSnap')

#-------------------------------------------------------------------------------
def OnStart():
  #Command starts here, all it does is display Action Panel.
  global props
  global node, surface, geoSet, topo
  global surface_normal
  
  cmd.ExecutionType = VC_COMMAND_TOOL
  
  #global inits
  node = None
  surface = None
  geoSet = None
  topo = None
  
  #-------------------------------------------------------------------------------
  # Define all the Properties to display on the Action Panel when command is launched
  #-------------------------------------------------------------------------------
  
  #for p in cmd.Properties:
  #  cmd.deleteProperty(p)
  
  if len(cmd.Properties) == 0:
    props = []
    props.append(addProperty( cmd, VC_BOOLEAN, 'UseCurrentSequence', True, False, False, False, updatePreview, ''))
    props.append(addProperty( cmd, VC_STRING, 'Sequence', True, False, False, 'Path', None, '' ))
    props.append(addStepProperty( cmd, VC_STRING, 'StatementType', True, False, False, 'LINEAR', updatePreview, '', ['PATH', 'LINEAR'] ))
    props.append(addLimitProperty( cmd, VC_REAL, 'OverShoot', True, False, False, 100., updatePreview, 'Distance', -500.00, 500.0 ))
    props.append(addProperty( cmd, VC_BOOLEAN, 'UseProcessSignal', True, False, False, True, updatePreview, ''))
    props.append(addLimitProperty( cmd, VC_INTEGER, 'ProcessSignal', True, False, False, 100, None, '', 0, 4096 ))
    props.append(addLimitProperty( cmd, VC_REAL, 'SignalOnDistance', True, False, False, 0., updatePreview, 'Distance', 0.00, 500.0 ))
    props.append(addLimitProperty( cmd, VC_REAL, 'SignalOffDistance', True, False, False, 0., updatePreview, 'Distance', -500, 500.0 ))
    props.append(addLimitProperty( cmd, VC_REAL, 'ProcessSpeed', True, False, False, 1000.0, None, 'Velocity', 1.00, 5000.0 ))
    props.append(addLimitProperty( cmd, VC_REAL, 'ApproachSpeed', True, False, False, 100.0, None, '', 1.00, 5000.0 ))
    props.append(addStepProperty( cmd, VC_STRING, 'ApproachMotion', True, False, False, 'PTP', None, '', ['PTP', 'LINEAR'] ))
    props.append(addLimitProperty( cmd, VC_REAL, 'IndexDistance', True, False, False, 100.0, updatePreview, 'Distance', 0.00, 1000.0 ))
    props.append(addStepProperty( cmd, VC_STRING, 'IndexDirection', True, False, False, 'X+', updatePreview, '', ['X+', 'X-', 'Y+', 'Y-'] ))
    props.append(addStepProperty( cmd, VC_STRING, 'SliceMethod', True, False, False, 'Linear', updatePreview, '', ['Linear','CenterLine'] ))
    props.append(addLimitProperty( cmd, VC_REAL, 'SliceOffset', True, False, False, 0.0, updatePreview, 'Distance', -1000.00, 1000.0 ))
    props.append(addLimitProperty( cmd, VC_REAL, 'SliceAngle', True, False, False, 0.0, updatePreview, 'Angle', -60.0, 60.0 ))
    props.append(addStepProperty( cmd, VC_STRING, 'Pattern', True, False, False, 'Zig-Zag', updatePreview, '', ['Zig-Zag', 'Zag-Zig', 'Zig-Zig', 'Zag-Zag'] ))
    props.append(addLimitProperty( cmd, VC_REAL, 'StartPercent', True, False, False, 0.0, updatePreview, 'Percentage', 0.0, 1.0))
    props.append(addLimitProperty( cmd, VC_REAL, 'EndPercent', True, False, False, 1.0, updatePreview, 'Percentage', 0.0, 1.0))
    props.append(addStepProperty( cmd, VC_STRING, 'ApproachAxis', True, False, False, 'Z+', updatePreview, '', ['Z+', 'Z-', 'X+', 'X-','Y+', 'Y-'] ))
    props.append(addStepProperty( cmd, VC_STRING, 'TravelAxis', True, False, False, 'X-', updatePreview, '', ['X+', 'X-', 'Y+', 'Y-'] ))
    props.append(addLimitProperty( cmd, VC_REAL, 'TravelAngle', True, False, False, 0.0, updatePreview, 'Angle', -90.00, 90.0 ))
    props.append(addLimitProperty( cmd, VC_REAL, 'RollAngle', True, False, False, 0.0, updatePreview, 'Angle', -90.00, 90.0 ))
    props.append(addStepProperty( cmd, VC_STRING, 'RollType', True, False, False, 'Constant', updatePreview, '', ['Constant', 'FollowPath'] ))
    props.append(addLimitProperty( cmd, VC_REAL, 'MinDistance', True, False, False,   50.0, updatePreview, 'Distance', 10., 500.0 ))
    props.append(addLimitProperty( cmd, VC_REAL, 'MaxDistance', True, False, False, 10000.0, updatePreview, 'Distance', 500., 10000.0 ))
    props.append(addLimitProperty( cmd, VC_REAL, 'ChordalDeviation', True, False, False, 1.0, updatePreview, 'Distance', 0.0, 25.0 ))
    props.append(addLimitProperty( cmd, VC_REAL, 'AngleDeviation', True, False, False, 5.0, updatePreview, 'Angle', 0.0, 45.0 ))
    props.append(addProperty( cmd, VC_BUTTON, 'Select Surface', True, False, False, True, selectSurface, '' ))
    props.append(addProperty( cmd, VC_BUTTON, 'Generate', True, False, False, True, generate, '' ))
    props.append(addProperty( cmd, VC_BUTTON, 'Cancel', True, False, False, True, cleanup, '' ))
  
  executeInActionPanel()
#-------------------------------------------------------------------------------
def generate(prop = None):
  #Generate path program
  global node, surface, geoSet, topo
  
  #addPaintPath(node,geoSet,topo,surface)
  if node:
    path = getPath(node,geoSet,topo,surface)
    createProgramStatements(node,path)
    cleanup()
  else:
    'No surface selected.'
#-------------------------------------------------------------------------------
def cleanup(prop = None):
  #Do cleanup
  
  path_comp = app.findComponent('PathPreview')
  if path_comp:
    app.deleteComponent(path_comp)
  app.render()
    
  jog_cmd = app.findCommand('interactiveTranslateTCP')
  jog_cmd.execute()
#-------------------------------------------------------------------------------
def selectSurface(prop = None):
  #Select surface from 3D model

  #Init topology picking react only to surface elements
  cmdPick.clear()
  cmdPick.SnapOnPoint = False
  cmdPick.SnapOnEdge = False
  cmdPick.SnapOnFrame = False
  cmdPick.SnapOnPolygon = True
  cmdPick.SnapOnSurface = True
  cmdPick.SnapOnNodeCenter = False
  cmdPick.SnapOnNodeOrigin = False

  # Use single topology pick mode
  cmdPick.ContinousMode = False

  # Register to get picking events
  cmdPick.OnTargetSet = OnPicked
  print 'Pick face to generate paint path'
  #cmdPick.execute()
  cmdPick.executeAsTool()
#-------------------------------------------------------------------------------
def OnPicked(cmdPick):
  #Topology picking event received, check if it contains valid result
  global node, surface, geoSet, topo, normal
  
  try:
    topotest = cmdPick.TargetGeoset.Topology
  except:
    print 'Add-on requries access to topology (Premium or Robotics license).'
    return
  
  node = cmdPick.TargetNode
  if node:
    #We are interested in the picked surface
    if cmdPick.TargetSurfaceAvailable:
      #Get curve basic information (notice, the topology is based on indices to maximize the performance)
      surface = cmdPick.TargetSurfaceIndex
      geoSet = cmdPick.TargetGeoset
      topo = geoSet.Topology
      updatePreview(None)
    else:
      print 'No surface found'
    #endif
  #endif

  #Remember to unregister the picking listener
  cmdPick.OnTargetSet = None
  app.render()
#-------------------------------------------------------------------------------
def updatePreview(prop):
  #Update preview of the path
  global node, surface, geoSet, topo
  
  #Update GUI also
  p = cmd.getProperty('TravelAxis')
  travel_axis = p.Value
  if 'Z' in cmd.ApproachAxis:
    travel_axes = ['X+', 'X-', 'Y+', 'Y-']
  elif 'X' in cmd.ApproachAxis:
    travel_axes = ['Z+', 'Z-', 'Y+', 'Y-']
  else:
    travel_axes = ['X+', 'X-', 'Z+', 'Z-']
  if not travel_axis in travel_axes:
    travel_axis = travel_axes[0]
  p.StepValues = travel_axes
  p.Value = travel_axis
  #
  p = cmd.getProperty('Sequence')
  p.WritableWhenConnected = not cmd.UseCurrentSequence
  p.WritableWhenDisconnected = not cmd.UseCurrentSequence
  #
  p = cmd.getProperty('ProcessSignal')
  p.WritableWhenConnected = cmd.UseProcessSignal
  p.WritableWhenDisconnected = cmd.UseProcessSignal
  p = cmd.getProperty('SignalOnDistance')
  p.WritableWhenConnected = cmd.UseProcessSignal
  p.WritableWhenDisconnected = cmd.UseProcessSignal
  p = cmd.getProperty('SignalOffDistance')
  p.WritableWhenConnected = cmd.UseProcessSignal
  p.WritableWhenDisconnected = cmd.UseProcessSignal
  
  if node:
    path = getPath(node,geoSet,topo,surface)
    visualizePath(node,path)
    print 'Point count: ', len(path)
#-------------------------------------------------------------------------------
def getPath(node, geoSet, topo, surface):
  #Get path as 2-tuple (Boolean processOn, vcMatrix positionInNode)
  global surface_normal
  
  #Get face bound in node coordinates
  if surface >= 0:
    center, diagonal = topo.getFaceBound(surface)
  else:
    center = vcMatrix.new()
    center.P = geoSet.BoundCenter
    diagonal = geoSet.BoundDiagonal
  #endif
  
  #Find surface coordinates, A is surface (~) mean normal, N is long side and O is short side, P is bound box center
  A = vcVector.new()
  curve_loop = topo.getFaceCurveLoop(surface, 0)
  curve_count = topo.getCurveLoopCurveCount(curve_loop)
  for i in range(curve_count):
    curve = topo.getCurveLoopCurve(curve_loop, i)
    curve_positions = topo.getCurvePositions(curve)
    for m in curve_positions:
      A = A + m.A
  A.normalize()
  surface_normal = vcVector.new(A)
  N = vcVector.new(1, 0, 0)
  if (A ^ N).length()  < 1e-6:
    if A.X > 0:
      N = vcVector.new(0,0,-1)
    else:
      N = vcVector.new(0,0,1)
  O = A ^ N
  O.normalize()
  N = O ^ A
  N.normalize()
  face_coords = vcMatrix.new()
  face_coords.N = N
  face_coords.O = O
  face_coords.A = A
  face_coords.P = center.P
  face_coords.uniform()
  iface_coords = vcMatrix.new(face_coords)
  iface_coords.invert()
  
  #Find diagonal in surface coordinates
  diagonal_in_face = vcVector.new(0,0,0)
  curve_loop = topo.getFaceCurveLoop(surface, 0)
  curve_count = topo.getCurveLoopCurveCount(curve_loop)
  for i in range(curve_count):
    curve = topo.getCurveLoopCurve(curve_loop, i)
    curve_positions = topo.getCurvePositions(curve)
    for m in curve_positions:
      p = iface_coords * m.P
      diagonal_in_face.X = max(abs(p.X), diagonal_in_face.X)
      diagonal_in_face.Y = max(abs(p.Y), diagonal_in_face.Y)
  
  if diagonal_in_face.Y > diagonal_in_face.X:
    #Y is long side, swap for X=long side
    face_coords.rotateRelZ(90)
    iface_coords = vcMatrix.new(face_coords)
    iface_coords.invert()
    diagonal_in_face = vcVector.new(diagonal_in_face.Y,diagonal_in_face.X,0)
    
  #Create slicing plane and select index direction
  #Slicing plane's Z is the slicing plane normal and "index direction".
  #Slicing plane's X is "tool's travel direction"
  #Slicing plane's Y is "selected surface's normal"
  plane = vcMatrix.new()
  if 'X' in cmd.IndexDirection:
    width = 2.0*diagonal_in_face.X
    plane.A = 1.0 * face_coords.N
    plane.N = 1.0 * face_coords.O
    plane.O = plane.A ^ plane.N
  elif 'Y' in cmd.IndexDirection:
    width = 2.0*diagonal_in_face.Y
    plane.A = 1.0 * face_coords.O
    plane.N = (-1.0) * face_coords.N
    plane.O = plane.A ^ plane.N
  #endif
  if '-' in cmd.IndexDirection: 
    plane.A = -1.0*plane.A
    plane.N = plane.O ^ plane.A
  #endif
  plane.uniform()
  
  #Slicing parameters
  minChord = cmd.MinDistance
  maxChord = cmd.MaxDistance
  chordalDeviation = cmd.ChordalDeviation
  angularDeviation = cmd.AngleDeviation
  cdsq = chordalDeviation*chordalDeviation
  cosAD = cos( radians(angularDeviation) )

  #Modify slicing plane
  if surface >= 0 and cmd.SliceMethod == 'CenterLine':
    #Center line slicing uses center line curve to create slicing planes instead of progressing linearly towards indexing direction.
    #Difference is shown best for non-linear surfaces
    centerPlane = vcMatrix.new(plane)
    centerPlane.P = center.P
    centerPlane.rotateRelY(-90)
    
    #Get centerline curve
    curves = topo.getFacePlaneIntersections( surface, centerPlane, minChord, maxChord, chordalDeviation, angularDeviation )
    
    #Check that curves have consistent direction
    if len(curves) > 0:
      p0 = curves[0][0].P
      if len(curves[0]) > 1:
        p1 = curves[0][-1].P
      else:
        p1 = curves[-1][-1].P
      #endif
    #endif
    if (p1-p0)*plane.A < -1.0: # oops, wrong way: fix it
      curves.reverse()
      for curve in curves: 
        curve.reverse() 
        for c in curve: 
          c.rotateRelZ(180)
      #endfor
    #endif

    #Append curve at start and end in case centerline doesn't cover entire bounding box
    centerLine = [vcMatrix.new(curves[0][0])]
    for curve in curves:
      for p in curve: 
        centerLine.append(p)
    #endfor
    centerLine.append( vcMatrix.new(centerLine[-1]) ) 
    p0 = centerLine[0].P
    p1 = centerLine[-1].P
    p0 = iface_coords * p0
    p1 = iface_coords * p1
    if cmd.IndexDirection == 'X+':
      p0.X = min(-diagonal_in_face.X,p0.X)
      p1.X = max(diagonal_in_face.X,p1.X)
    elif cmd.IndexDirection == 'X-':
      p0.X = max(diagonal_in_face.X,p0.X)
      p1.X = min(-diagonal_in_face.X,p1.X)
    elif cmd.IndexDirection == 'Y+':
      p0.Y = min(-diagonal_in_face.Y,p0.Y)
      p1.Y = max(diagonal_in_face.Y,p1.Y)
    elif cmd.IndexDirection == 'Y-':
      p0.Y = max(diagonal_in_face.Y,p0.Y)
      p1.Y = min(-diagonal_in_face.Y,p1.Y)
    #endif
    p0 = face_coords * p0
    p1 = face_coords * p1
    centerLine[0].P = p0
    centerLine[-1].P = p1
    
    #Get width in indexing direction
    width = 0.0
    for i in range(1,len(centerLine)): 
      width += (centerLine[i].P-centerLine[i-1].P).length()
    start = cmd.SliceOffset
    start += cmd.StartPercent * width
    width *= abs(cmd.EndPercent - cmd.StartPercent)

    #Calculate step size and number of steps
    stepSize = width/round(width/cmd.IndexDistance)
    numSteps = int(width/stepSize)

    #Get first slicing plane
    plane.O, plane.P = getCurvePosition( centerLine, start + stepSize/2. )
    A = plane.N ^ plane.O
    A.normalize()
    O = A ^ plane.N
    O.normalize()
    plane.O = O
    plane.A = A
    plane.rotateRelX( cmd.SliceAngle )
  else:
    #Linear slicing
    
    #Get width in slicing direction
    width /= cos( radians(cmd.SliceAngle) )
    full_width = width
    start = cmd.SliceOffset
    start += cmd.StartPercent * width
    width *= abs(cmd.EndPercent - cmd.StartPercent)
    
    #Calculate step size and number of steps
    stepSize = width/round(width/cmd.IndexDistance)
    numSteps = int(width/stepSize)
    
    #Get first slicing plane
    plane.O = plane.A ^ plane.N
    plane.rotateRelX( cmd.SliceAngle )
    plane.P = center.P - (full_width/2.0 - start - stepSize/2.0)*plane.A
    #endif
  #endif
  
  #Switch initial sign for Zag
  if cmd.Pattern in ['Zag-Zig','Zag-Zag']:
    sign = -1.0
  else:
    sign = 1.0
  #endif
  init_sign = sign

  #Do the slicing
  path = []
  s = None
  process_on = False
  for step in range(numSteps):
    newPass = True

    if surface >= 0:
      #Should come here
      curves = topo.getFacePlaneIntersections(surface, plane, minChord, maxChord, chordalDeviation, angularDeviation)
    else:
      #Backup strategy
      curves = []
      for face in range(topo.FaceCount):
        curves.extend(topo.getFacePlaneIntersections(face, plane, minChord, maxChord, chordalDeviation, angularDeviation))
      #endfor
      
      stitchedCurves = []
      while curves:
        curve = curves.pop()
        p1 = curve[0].P
        p2 = curve[-1].P
        i = 0
        while i < len(curves):
          testCurve = curves[i]
          q1 = testCurve[0].P
          q2 = testCurve[-1].P
          if (p2-q1).length() < 0.5:
            curve.pop()
            curve.extend(testCurve)
            p2 = q2
            curves.pop(i)
            i = 0
          elif (p2-q2).length() < 0.5:
            testCurve.reverse()
            curve.pop()
            curve.extend(testCurve)
            p2 = q1
            curves.pop(i)
            i = 0
          elif (p1-q2).length() < 0.5:
            curve.pop(0)
            testCurve.extend(curve)
            curve = testCurve
            p1 = q1
            curves.pop(i)
            i = 0
          elif (p1-q1).length() < 0.5:
            testCurve.reverse()
            curve.pop(0)
            testCurve.extend(curve)
            curve = testCurve
            p1 = q2
            curves.pop(i)
            i = 0
          else:
            i += 1
          #endif
        #endwhile
        stitchedCurves.append(curve)
      #endwhile
      curves = stitchedCurves
    #endif
    
    #Get next slicing plane
    if surface >= 0 and cmd.SliceMethod == 'CenterLine':
      plane.O, plane.P = getCurvePosition( centerLine, start + (step+1)*stepSize+stepSize/2. )
      A = plane.N ^ plane.O
      A.normalize()
      O = A ^ plane.N
      O.normalize()
      plane.O = O
      plane.A = A
      plane.rotateRelX( cmd.SliceAngle )
    else:
      plane.P = plane.P + stepSize*plane.A
    #endif

    if not curves: 
      #No curves this step for some reason
      continue
    
    #Check that curves have consistent direction
    if len(curves) > 0:
      p0 = curves[0][0].P
      if len(curves[0]) > 1:
        p1 = curves[0][-1].P
      else:
        p1 = curves[-1][-1].P
      #endif
    #endif
    if (p1-p0)*plane.N < -1.0:
      curves.reverse()
      for curve in curves: 
        curve.reverse() 
        for c in curve: 
          c.rotateRelZ(180)
      #endfor
    #endif
    
    #Repair curve travel direction (may get broken at certain situations)
    all_points = []
    for points in curves:
      all_points.extend(points)
    #endfor
    for i in range(len(all_points)):
      p0 = all_points[i]
      p1 = None
      #Try get travel direction from ahead
      for j in range(i+1,len(all_points)):
        if (p0.P-all_points[j].P).length() > minChord:
          #Can be used for travel direction
          p1 = all_points[j]
          break
        #endif
      #endfor
      if p1:
        N = p1.P-p0.P
        N.normalize()
        O = p0.A ^ N
        O.normalize()
        N = O ^ p0.A
        N.normalize()
        p0.N = N
        p0.O = O
      else:
        #Try get travel direction from ahead
        for j in range(i-1,-1,-1):
          if (p0.P-all_points[j].P).length() > minChord:
            #Can be used for travel direction
              p1 = all_points[j]
              break
        if p1:
          N = p0.P-p1.P
          N.normalize()
          O = p0.A ^ N
          O.normalize()
          N = O ^ p0.A
          N.normalize()
          p0.N = N
          p0.O = O
        #endif
      #endif
    #endfor
    
    #Reverse curves for Zag
    if sign < 0.0:
      curves.reverse()
      for curve in curves: 
        curve.reverse() 
        for p in curve:
          p.rotateRelZ(180)
    #endif
    
    #Filter points
    if not (surface >= 0) or topo.getFaceType(surface) == VC_FACE_TESSELATED:
      for i, points in enumerate(curves):
        prevP = points[0]
        prevJ = 0
        filteredPoints = [prevP]
        numPoints = len(points)
        j = 1
        skippedPoint = None
        while j < numPoints:
          p = points[j]
          dp = (p.P - prevP.P)
          length = dp.length()
          if minChord and length < minChord: 
            j += 1
          elif chordalDeviation or angularDeviation:
            v2 = p.P - prevP.P
            for k in range( prevJ+1, j ):  
              v1 = points[k].P - prevP.P
              if (chordalDeviation and v1*v1 - (v1*v2)*(v1*v2)/(v2*v2) > cdsq) or (angularDeviation and prevP.A*points[k].A < cosAD):
                prevP = points[j-1]
                prevJ = j-1
                filteredPoints.append(prevP)
                break
              #endif
            else:
              #No break
              j += 1
            #endfor
          else:
            filteredPoints.append(p)
            prevP = p
            j += 1
        #endwhile
        
        if filteredPoints[-1] != points[-1]:
          filteredPoints.append(points[-1])
        #endif
        curves[i] = filteredPoints
      #endfor
    #endif
    
    
    #For negative overshoot move start/end points inwards
    all_points = []
    for points in curves:
      all_points.extend(points)
    if cmd.OverShoot < 0.0 and len(all_points) >= 2:
      p0 = all_points[0]
      p1 = all_points[1]
      p0_ = all_points[-1]
      p1_ = all_points[-2]
      if len(all_points) == 2:
        shift = min(abs(cmd.OverShoot),0.49*(p1.P-p0.P).length())
        p0.translateRel(shift, 0., 0.)
        p0_.translateRel(-shift, 0., 0.)
      else:
        shift = min(abs(cmd.OverShoot),0.99*(p1.P-p0.P).length())
        p0.translateRel(shift, 0., 0.)
        shift = min(abs(cmd.OverShoot),0.99*(p1_.P-p0_.P).length())
        p0_.translateRel(-shift, 0., 0.)
    
    #Start overshoot point
    if cmd.OverShoot > 0.0:
      m = vcMatrix.new(curves[0][0])
      m.translateRel( -cmd.OverShoot, 0., 0. )
      m = setOrientation(m, sign != init_sign and cmd.RollType == 'Constant')
      path.append([process_on,m])
    #endif

    #First signal-on point
    if cmd.SignalOnDistance >= cmd.OverShoot:
      process_on = True
    elif cmd.SignalOnDistance > 0.0:
      m = vcMatrix.new(curves[0][0])
      m.translateRel( -cmd.SignalOnDistance, 0., 0. )
      m = setOrientation(m, sign != init_sign and cmd.RollType == 'Constant')
      path.append([process_on,m])
      process_on = True
    #endif
    
    i = 0
    hole_l = 0
    for points in curves:
      if i > 0:
        #A new curve after a hole, create new signal-on point
        #hole_l = (p.P-points[0].P).length()
        if cmd.SignalOnDistance > 0 and cmd.SignalOnDistance + cmd.SignalOffDistance < hole_l:
          shift = cmd.SignalOnDistance
          m = vcMatrix.new(points[0])
          m.translateRel(-shift, 0., 0. )
          m = setOrientation(m, sign != init_sign and cmd.RollType == 'Constant')
          path.append([process_on,m])
          process_on = True
        #endif
      
      #Actual process points
      for p in points:
        m = vcMatrix.new(p)
        m = setOrientation(m, sign != init_sign and cmd.RollType == 'Constant')
        path.append([process_on,m])
        process_on = True
      #endfor
      
      #A hole, create new signal-off point
      if i < len(curves) - 1:
        hole_l = (p.P-curves[i+1][0].P).length()
        if cmd.SignalOffDistance > 0 and cmd.SignalOffDistance + cmd.SignalOnDistance < hole_l:
          shift = cmd.SignalOffDistance
          m = vcMatrix.new(p)
          m.translateRel(shift, 0., 0.)
          m = setOrientation(m, sign != init_sign and cmd.RollType == 'Constant')
          path.append([process_on,m])
          process_on = False
        elif cmd.SignalOffDistance + cmd.SignalOnDistance < hole_l:
          process_on = False
          
      i += 1  
    #endfor
    
    #Last signal-off point
    if cmd.SignalOffDistance == 0.0:
      process_on = False
    elif cmd.SignalOffDistance < cmd.OverShoot:
      m = vcMatrix.new(p)
      m.translateRel( cmd.SignalOffDistance, 0., 0. )
      m = setOrientation(m, sign != init_sign and cmd.RollType == 'Constant')
      path.append([process_on,m])
      process_on = False
    #endif

    #End overshoot point
    if cmd.OverShoot > 0.0:
      m = vcMatrix.new(p)
      m.translateRel( cmd.OverShoot, 0., 0. )
      m = setOrientation(m, sign != init_sign and cmd.RollType == 'Constant')
      path.append([process_on,m])
      if cmd.SignalOffDistance < cmd.OverShoot: 
        process_on = False
    #endif
    
    if cmd.Pattern in ['Zig-Zag','Zag-Zig']: sign *= -1.0
  #endfor (slicing loop)

  return path
#-------------------------------------------------------------------------------
def setOrientation(m, flip_roll):
  #Raw matrix has Z- as approach axis and X+ as travel axis.
  #Set roll and travel angles and then change approach and travel axes.
  
  m.rotateRelY(-cmd.TravelAngle)
  m.rotateRelZ(cmd.RollAngle)
  if flip_roll:
    m.rotateRelZ(180)
  
  n = None
  o = None
  a = None
  
  if cmd.ApproachAxis == 'Z+':
    a = (-1.0) * m.A
  elif cmd.ApproachAxis == 'Z-':
    a = (1.0) * m.A
  elif cmd.ApproachAxis == 'X+':
    n = (-1.0) * m.A
  elif cmd.ApproachAxis == 'X-':
    n = (1.0) * m.A
  elif cmd.ApproachAxis == 'Y+':
    o = (-1.0) * m.A
  else:#if cmd.ApproachAxis == 'Y-':
    o = (1.0) * m.A
  
  if cmd.TravelAxis == 'Z+':
    a = (1.0) * m.N
  elif cmd.TravelAxis == 'Z-':
    a = (-1.0) * m.N
  elif cmd.TravelAxis == 'X+':
    n = (1.0) * m.N
  elif cmd.TravelAxis == 'X-':
    n = (-1.0) * m.N
  elif cmd.TravelAxis == 'Y+':
    o = (1.0) * m.N
  else:#if cmd.TravelAxis == 'Y-':
    o = (-1.0) * m.N
    
  if n and o:
    a = n ^ o
  elif n and a:
    o = a ^ n
  elif o and a:
    n = o ^ a
  else:
    return m
  m.N = n
  m.O = o
  m.A = a
  m.uniform()
  
  return m
#-------------------------------------------------------------------------------
def visualizePath(node, path):
  #Visualize path as a component with line sets
  
  if not path:
    return
  
  path_comp = app.findComponent('PathPreview')
  if path_comp:
    app.deleteComponent(path_comp)
  path_comp = app.createComponent()
  path_comp.Name = 'PathPreview'
  
  node.attach(path_comp, True)
  
  feat = path_comp.RootFeature.createFeature(VC_GEOMETRY, 'Path')
  
  #Path visu
  process_lineset = feat.Geometry.createGeometrySet(VC_COMPACTLINESET)
  via_lineset = feat.Geometry.createGeometrySet(VC_COMPACTLINESET)
  
  process_lineset.Material =  app.findMaterial('deep_pink')
  via_lineset.Material = app.findMaterial('light_cyan')
  process_lineset.LineWidth = 3
  via_lineset.LineWidth = 3
  
  line = []
  prev_point = None
  prev_m = None
  m0 = None
  for point in path:
    #point[0]: process_on
    #point[1]: position matrix in node
    #
    #Do some offset in approach direction
    m = vcMatrix.new(point[1])
    #offset = 1
    #if cmd.ApproachAxis == 'Z+':
    #  m.translateRel(0,0,-offset)
    #elif cmd.ApproachAxis == 'Z-':
    #  m.translateRel(0,0,offset)
    #elif cmd.ApproachAxis == 'X+':
    #  m.translateRel(-offset,0,0)
    #elif cmd.ApproachAxis == 'X-':
    #  m.translateRel(offset,0,0)
    #elif cmd.ApproachAxis == 'Y+':
    #  m.translateRel(0,-offset,0)
    #elif cmd.ApproachAxis == 'Y-':
    #  m.translateRel(0,offset,0)
    #
    if prev_point and point[0] != prev_point[0]:
      if prev_point[0] or not cmd.UseProcessSignal:
        process_lineset.addLine(line)
        #Add arrows
        for i in range(len(line)-1):
          p0 = line[i]
          p1 = line[i+1]
          arrow = getArrow(p0,p1)
          if arrow:
            process_lineset.addLine(arrow)
      else:
        via_lineset.addLine(line)
      line = []
      line.append(prev_m.P)
    line.append(m.P)
    prev_point = point
    prev_m = m
    if not m0:
      m0 = m
    
  if prev_point:
    if prev_point[0] or not cmd.UseProcessSignal:
      process_lineset.addLine(line)
      #Add arrows
      for i in range(len(line)-1):
        p0 = line[i]
        p1 = line[i+1]
        arrow = getArrow(p0,p1)
        if arrow:
          process_lineset.addLine(arrow)
    else:
      via_lineset.addLine(line)
  
  #Tool frame visu
  x_lineset = feat.Geometry.createGeometrySet(VC_COMPACTLINESET)
  y_lineset = feat.Geometry.createGeometrySet(VC_COMPACTLINESET)
  z_lineset = feat.Geometry.createGeometrySet(VC_COMPACTLINESET)
  
  x_lineset.Material =  app.findMaterial('red')
  y_lineset.Material = app.findMaterial('green')
  z_lineset.Material = app.findMaterial('blue')
  x_lineset.LineWidth = 5
  y_lineset.LineWidth = 5
  z_lineset.LineWidth = 5
  
  
  line = [m0.P, m0.P + 100. * m0.N]
  x_lineset.addLine(line)
  line = [m0.P, m0.P + 100. * m0.O]
  y_lineset.addLine(line)
  line = [m0.P, m0.P + 100. * m0.A]
  z_lineset.addLine(line)
  
  path_comp.rebuild()
  
  app.render()
#-------------------------------------------------------------------------------
def getArrow(p0, p1):
  global surface_normal
  #Get arrowhead for midpoint in line p0->p1
  
  arrow_length = 50.0
  arrow = []
  n = p1-p0
  if n.length() < 2 * arrow_length:
    return arrow
  p = p0 + (0.5*n)
  n.normalize()
  if surface_normal:
    a = vcVector.new(surface_normal)
  else:
    a = vcVector.new(0,0,1)
    if (a ^ n).length() < 1e-6:
      a = vcVector.new(1,0,0)
  o = a ^ n
  o.normalize()
  a = n ^ o
  m = vcMatrix.new()
  m.N = n
  m.O = o
  m.A = a
  m.P = p
  m.uniform()
  m.rotateRelZ(-20)
  arrow.append(p + (-arrow_length*m.N))
  arrow.append(p)
  m.rotateRelZ(40)
  arrow.append(p + (-arrow_length*m.N))
  #arrow.append(p)
  return arrow
#-------------------------------------------------------------------------------
def getCurvePosition(curve, distance):
  #Return curve normal and position for given distance
  dist = 0.0
  for i in range(1, len(curve)):
    dist0 = dist
    dp = curve[i].P-curve[i-1].P
    dist += dp.length()
    if dist > distance:
      m = vcMatrix.new(curve[i-1])
      m.P = m.P + ((distance-dist0)/(dist-dist0))*dp 
      return  m.A, m.P
    #endif
  #endfor
  return curve[-1].A, curve[-1].P
#-------------------------------------------------------------------------------
def createProgramStatements(node, path):
  #Generate path into robot program
  global process_on
  global scope
  global controller

  robot = app.TeachContext.ActiveRobot
  executor = robot.findBehavioursByType(VC_ROBOTEXECUTOR)[0]
  controller = executor.Controller
  program = executor.Program
  
  routine = None
  if cmd.UseCurrentSequence:
    routine = app.TeachContext.ActiveRoutine
  if not routine:
    routine = program.findRoutine(cmd.Sequence)
    if not routine: 
      routine = program.addRoutine(cmd.Sequence)
    routine.clear()
    #
    stmt = None
    if len(program.MainRoutine.Statements) > 0 and program.MainRoutine.Statements[-1].Type == VC_STATEMENT_CALL and program.MainRoutine.Statements[-1].Routine == routine: 
      pass
    else:
      stmt = program.MainRoutine.addStatement(VC_STATEMENT_CALL, -1 )
      stmt.Routine = routine
      
  scope = routine
  
  target = controller.createTarget()

  for tool in controller.Tools:
    if tool.Name == target.ToolName: 
      break
  else:
    tool = None
  #endfor

  for base in controller.Bases:
    if base.Name == target.BaseName: 
      if base.Node:
        BiW = base.Node.WorldPositionMatrix*base.PositionMatrix
      else:
        BiW = controller.Component.WorldPositionMatrix * controller.WorldTransformMatrix * base.PositionMatrix
      #endif
      break
    #endif
  else:
    base = None
    BiW = controller.Component.WorldPositionMatrix * controller.WorldTransformMatrix
  #endfor

  WiB = vcMatrix.new(BiW)
  WiB.invert()
  PiB = WiB * node.WorldPositionMatrix
  NiW = node.WorldPositionMatrix

  rootNode = controller.Component
  while rootNode.Parent != rootNode.World:
    rootNode = rootNode.Parent
  WiR = rootNode.InverseWorldPositionMatrix * node.WorldPositionMatrix

  s = None
  motion = cmd.ApproachMotion
  speed = cmd.ApproachSpeed
  continuous = False
  process_on = False
  i = 0
  
  if cmd.StatementType == 'PATH':
    #Create process path handler if path is used
    handler = createProcessPathHandler(executor)
    addPath(PiB, path, base, tool, handler)
  else:
    for point in path:
      new_process_on = point[0]
      m = point[1]
      
      if i > 0:
        addProcessSignal(new_process_on)
      
      if i < len(path)-1:
        if path[i][0] != path[i+1][0] and cmd.ProcessSignal >= 0:
          #Process signal after this point, use fine point
          continuous = False
      #endif
          
      s = addMove(NiW, m, motion, base, tool, speed, continuous )
      
      continuous = True
      motion = "LIN"
      speed = cmd.ProcessSpeed
      
      i += 1
    #endfor
    addProcessSignal(False)
  #endif

  app.render()
#-------------------------------------------------------------------------------
def createProcessPathHandler(exe):
  #Create path statement handler
  
  comp = exe.Component
  handler = comp.getBehaviour('ProcessPathHandler')
  if not handler:
    handler = comp.createBehaviour(VC_PYTHONPROCESSHANDLER,'ProcessPathHandler')
  
  folder = os.path.dirname(cmd.getCommandPath())[8:] + '\\'
  file = open(folder + 'ProcessPathHandler.py', 'r')
  handler.Script = file.read()
  file.close()
  
  return handler
#-------------------------------------------------------------------------------
def addProcessSignal(value):
  global process_on

  if process_on != value and cmd.UseProcessSignal: 
    stmt = addSetBin(cmd.ProcessSignal, value)
  else: 
    stmt = None

  process_on = value
  return stmt
#-------------------------------------------------------------------------------
def addPath(PiB, path, base, tool, handler):
  stmt = scope.addStatement(VC_STATEMENT_PATH, -1)
  stmt.Base = base
  stmt.Tool = tool
  stmt.ProcessHandler = handler
  stmt.clearPositions()
  prop = stmt.createProperty(VC_INTEGER, 'ProcessSignal')
  if cmd.UseProcessSignal:
    prop.Value = cmd.ProcessSignal
  else:
    prop.Value = -1
  prop = stmt.createProperty(VC_BOOLEAN, 'ApproachMotionPtp')
  prop.Value = (cmd.ApproachMotion == 'PTP')
  prop = stmt.createProperty(VC_REAL, 'ApproachSpeed')
  prop.Value = cmd.ApproachSpeed
  
  #add schema properties/fields for table 
  my_props = [VC_MATRIX, 'Position',\
              VC_BOOLEAN, 'ProcessPoint',\
              VC_REAL, 'MaxSpeed',\
              VC_BOOLEAN, 'IsReference',\
              VC_BOOLEAN, 'IsContinuous',\
              VC_REAL, 'AccuracyValue',\
              VC_INTEGER, 'AccuracyMethod',\
              VC_REAL, 'Acceleration',\
              VC_REAL, 'Deceleration']
  prop_indices = stmt.addSchemaProperties(my_props) 
    
  #add data to path table 
  stmt.beginBatchUpdate(len(path)) 
  for point_index in range(len(path)):
    position = PiB * path[point_index][1]
    
    process_point = True
    if cmd.UseProcessSignal:
      process_point = path[point_index][0]
      if not process_point and point_index < len(path) - 1 and path[point_index + 1][0]:
        #Process start point
        process_point = True
      
    speed = cmd.ProcessSpeed
    if point_index == 0:
      speed = max(cmd.ApproachSpeed,cmd.ProcessSpeed)
    isRef = False
    if point_index == 0 or point_index == len(path)-1:
      isRef = True
    isContinuous = True
    accu = min(cmd.MinDistance/2, cmd.IndexDistance)
    acc = controller.MaxCartesianAccel
    position_values = (position, process_point, speed, isRef, isContinuous, accu, VC_MOTIONTARGET_AM_DISTANCE, acc, acc)
    stmt.setSchemaValues(point_index, prop_indices, position_values)
  stmt.endBatchUpdate(len(path)) 
  
  return stmt
#-------------------------------------------------------------------------------
def addMove(NiW, m, motype, base, tool, speed, flyby = False):

  m.uniform()
  worldPosition = NiW * m

  if motype == 'PTP':
    s = addPTPStatement( worldPosition, base, tool, min(speed,100.0)/100.0, flyby )
  else:
    s = addLINStatement( worldPosition, base, tool, speed, flyby )
#-------------------------------------------------------------------------------
def addPTPStatement(worldPosition, base, tool, speed, flyby = False):
  stmt = scope.addStatement(VC_STATEMENT_PTPMOTION, -1)
  stmt.Base = base
  stmt.Tool = tool
  stmt.JointSpeed = speed
  if flyby:
    #stmt.Acceleration = 0
    #stmt.Deceleration = 0
    stmt.AccuracyMethod = VC_MOTIONTARGET_AM_DISTANCE
    stmt.AccuracyValue = min(cmd.MinDistance/2, cmd.IndexDistance)
  #endif
  positionFrame = stmt.Positions[0]
  positionFrame.PositionInWorld = worldPosition
  return stmt
#-------------------------------------------------------------------------------
def addLINStatement(worldPosition, base, tool, speed, flyby = False):
  stmt = scope.addStatement(VC_STATEMENT_LINMOTION, -1)
  stmt.Base = base
  stmt.Tool = tool
  stmt.MaxSpeed = speed
  if flyby:
    #stmt.Acceleration = 0
    #stmt.Deceleration = 0
    stmt.AccuracyMethod = VC_MOTIONTARGET_AM_DISTANCE
    stmt.AccuracyValue = min(cmd.MinDistance/2, cmd.IndexDistance)
  #endif

  positionFrame = stmt.Positions[0]
  positionFrame.PositionInWorld = worldPosition
  return stmt
#-------------------------------------------------------------------------------
def addSetBin(port, value):

  stmt = scope.addStatement(VC_STATEMENT_SETBIN, -1)
  stmt.OutputPort = port
  stmt.OutputValue = value
  return stmt 
#-------------------------------------------------------------------------------